#line 2111 "ASTGRC.nw"
#include "ASTGRC.hpp"
#include <cstdio>

namespace ASTGRC {
  
#line 499 "ASTGRC.nw"
GrcSynth::GrcSynth(Module *m, CompletionCodes &c)
  : module(m), code(c),
    surface_context(code.max() + 1),
    depth_context(code.max() +1),
    surface(surface_context, *this), depth(depth_context, *this),
    seltree(*this)
{
  assert(m);
  assert(m->types);
  integer_type = dynamic_cast<BuiltinTypeSymbol*>(m->types->get("integer"));
  assert(integer_type);
  boolean_type = dynamic_cast<BuiltinTypeSymbol*>(m->types->get("boolean"));
  assert(boolean_type);

  for ( SymbolTable::const_iterator i = m->signals->begin() ;
        i != m->signals->end() ; i++ ) {
    SignalSymbol *s = dynamic_cast<SignalSymbol*>(*i);
    assert(s);
    clone.sameSig(s);
  }
}
#line 2116 "ASTGRC.nw"
  
#line 632 "ASTGRC.nw"
GrcWalker::GrcWalker(Context &c, GrcSynth &e)
    : context(c), environment(e), clone(e.clone) {}
#line 637 "ASTGRC.nw"
STNode *GrcWalker::stnode(const ASTNode &n) {
  assert(environment.ast2st.find(&n) != environment.ast2st.end() );
  return environment.ast2st[&n];
}
#line 2117 "ASTGRC.nw"
  
#line 720 "ASTGRC.nw"
Status Surface::visit(Pause &s) {
  Enter *en = new Enter(stnode(s));
  assert(en->st);
  *en >> context(1);
  context(0) = en;
  return Status();
}
#line 759 "ASTGRC.nw"
Status Surface::visit(Exit &s) {
  assert(s.trap);
  context(0) = context(environment.code[s.trap]);
  push_onto(context(0), new Action(clone(&s)));
  return Status();
}
#line 844 "ASTGRC.nw"
Status Surface::visit(IfThenElse &s) {
  Enter *en;
  assert(s.predicate);
  Test *t = new Test(stnode(s), clone(s.predicate));
  *t >> ( (s.else_part != 0) ? recurse(s.else_part) : context(0))
     >> ( (s.then_part != 0) ? recurse(s.then_part) : context(0));
  context(0) = t;
  en = new Enter(stnode(s));
  push_onto(context(0), en);
  return Status();
}
#line 905 "ASTGRC.nw"
Status Surface::visit(StatementList &s) {

  for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	i != s.statements.rend() ; i++ ) {
    assert(*i);
    context(0) = synthesize(*i);
  }

  push_onto(context(0), new Enter(stnode(s)));
  return Status();
}
#line 970 "ASTGRC.nw"
Status Surface::visit(Loop &s) {
  context(0) = synthesize(s.body);
  Enter *en = new Enter(stnode(s));
  push_onto(context(0), en);
  return Status();
}
#line 1017 "ASTGRC.nw"
Status Surface::visit(Repeat &s) {
  context(0) = synthesize(s.body);
  StartCounter *stcnt = new StartCounter(s.counter, stnode(s));
  push_onto(context(0), new Action(stcnt));
  Enter *en = new Enter(stnode(s));
  push_onto(context(0), en);
  return Status();
}
#line 1084 "ASTGRC.nw"
Status Surface::visit(Suspend &s)
{
  assert(s.body);
  GRCNode *start = recurse(s.body);

  assert(stnode(s)->children.size() >= 1); // Should have at least one child
  STNode *bodytree = stnode(s)->children.front();
  assert(bodytree); // First child is STref for the body

  // Put an Enter for the suspend's body just before the code for the body
  push_onto(start, new Enter(bodytree));
   
  Delay *d = dynamic_cast<Delay*>(s.predicate);
  if (d) {
    if (d->is_immediate) {

      // An immediate predicate (e.g., suspend .. when immediate A)
      STNode *imleaf = stnode(s)->children.back();
      Enter *enimleaf = new Enter(imleaf);
      *enimleaf >> context(1); // Enter the immediate additional leaf
      Test *tst = new Test(bodytree, clone(d->predicate));
      *tst >> start >> enimleaf;
      start = tst;

    } else {

      // A counted suspend (e.g., suspend .. when 5 A)
      StartCounter *scnt = new StartCounter(d->counter, bodytree);
      push_onto(start, new Action(scnt));

    }
  }

  // Put an Enter for the suspend statement itself at the beginning
  push_onto(start, new Enter(stnode(s)));

  context(0) = start;
  return Status();
}
#line 1241 "ASTGRC.nw"
Status Surface::visit(Abort &s) {
  if (s.is_weak) throw IR::Error("weak abort.  Did the dismantler run?");

  // Synthesize the surface of the body

  context.push();
  assert(s.body);
  GRCNode *start = recurse(s.body);
  context.pop();

  // Add an enter node for STref node under the abort

  assert(stnode(s)->children.size() >= 1); // Should be at least one child
  assert(stnode(s)->children.front()); // First child is STref for this body

  // The selection tree node for the body of the abort
  STNode *bodytree = stnode(s)->children.front();

  push_onto(start, new Enter(bodytree));

  // Add a check for each immediate predicate and "initialize counter"
  // for each counted predicate
   
  for ( vector<PredicatedStatement*>::reverse_iterator i = s.cases.rbegin() ;
	i != s.cases.rend() ; i++ ) {
    assert(*i);
    assert((*i)->predicate);
    Delay *d = dynamic_cast<Delay*>((*i)->predicate);
    if (d) {
      if (d->is_immediate) {

	// An immediate predicate: add a test an a handler

	assert(d->counter == NULL); // immediate delays shoudn't be counted
	assert(d->predicate);
	// FIXME: does the Test need this reference to the abort STref node?
	Test *tst = new Test( bodytree, clone(d->predicate) );
	assert(tst->st);
	*tst >> start
	  // If the predicate has a body, send control there
	     >> ( ((*i)->body) ? recurse((*i)->body)
		  : context(0) );
	start = tst;

      } else {

	// A counted predicate: add code that initializes the counter

	push_onto(start, new Action(new StartCounter(d->counter, bodytree)));
	  
      }
    }
  }
 
  // Topmost node in the surface is an enter for the whole abort
  push_onto(start, new Enter(stnode(s)));

  context(0) = start;

  return Status();
}
#line 1428 "ASTGRC.nw"
Status Surface::visit(ParallelStatementList &s) {
  Fork *fork = new Fork();
  Sync *sync = new Sync(stnode(s));
  Terminate *t;
  int nthr;

  GRCNode **outer = context.continuations.top();
  assert(outer);
  context.push();

  // Create a new terminate for every possible exit level
  // and link each from the sync node
  
  // Synthesize each thread's surface
  for ( vector<Statement*>::iterator i = s.threads.begin() ;i != s.threads.end() ; i++ ) {
    assert(*i);
    nthr=i-s.threads.begin();

    for(int tl=0; tl<context.size; tl++){
	    t = new Terminate(tl, nthr);
	    *t >> sync;
	    context(tl)=t;

	    if(tl == 0){ // this is the self looping enter
	      Enter *en = new Enter( stnode(s)->children[nthr]->children[0] );
	      assert(en->st);
	      push_onto(context(tl), en);
            }
    }
   
    *fork >> recurse(*i); // it links thread to terminates, but each thread should have its own "enter"
  }

  // Connect each Terminate node with predecessors (i.e., that was
  // used by the threads) to the Sync and delete the rest.

  for ( int i = 0 ; i < context.size ; i++ ){
	*sync >> outer[i];
  }

  context.pop();
  context(0) = fork;
  push_onto(context(0), new Enter(stnode(s)));
  return Status();
}
#line 1590 "ASTGRC.nw"
Status Surface::visit(Trap &s) {

  assert(s.symbols);
  assert(s.symbols->begin() != s.symbols->end());
  SignalSymbol *ts = dynamic_cast<SignalSymbol*>((*(s.symbols->begin())));
  assert(ts);

  for (SymbolTable::const_iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++ ) {
    SignalSymbol *ss = dynamic_cast<SignalSymbol*>(*i);
    assert(ss);
    assert(ss->kind == SignalSymbol::Trap);
    clone.cloneLocalSignal(ss, environment.module->signals);
  }

  int level = environment.code[ts];
  assert(level > 1); // Should have been assigned by the dismantler

  // Esterel permits multiple handlers, but the dismantler should
  // have removed them
  if (s.handlers.size() > 1)
    throw IR::Error("Multiple trap handler.  Did the dismantler run?");

  // Surface for the handler is either a normal termination or
  // the handler for the body
  assert( s.handlers.empty() || s.handlers.front() );
  GRCNode *handlerSurface =
     s.handlers.empty() ? context(0) : recurse(s.handlers.front()->body);

  // Synthesize the body
  context.push();

  assert(handlerSurface);
  context(level) = handlerSurface;

  assert(s.body);
  GRCNode *surface = synthesize(s.body);

  context.pop();

  assert(stnode(s));
  assert(stnode(s)->children.front());
  push_onto(surface, new Enter(stnode(s)->children.front()));

  // Add "DefineSignal" nodes for each of the traps

  for (SymbolTable::const_iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++ ) {
    SignalSymbol *ss = dynamic_cast<SignalSymbol*>(*i);
    assert(ss);
    assert(ss->kind == SignalSymbol::Trap);
    push_onto(surface, new DefineSignal(clone(ss)));
    clone.clearSig(ss);
  }

  push_onto(surface, new Enter(stnode(s)));

  context(0) = surface;

  return Status();
}
#line 1753 "ASTGRC.nw"
Status Surface::visit(Signal &s) {
  for ( SymbolTable::const_iterator i = s.symbols->begin() ;
        i != s.symbols->end() ; i++ ) {
    SignalSymbol *sig = dynamic_cast<SignalSymbol*>(*i);
    assert(sig);
    clone.cloneLocalSignal(sig, environment.module->signals);
  }

  context(0) = synthesize(s.body);

  for ( SymbolTable::const_iterator i = s.symbols->begin() ;
        i!= s.symbols->end() ; i++ ) {
    SignalSymbol *sig = dynamic_cast<SignalSymbol*>(*i);
    assert(sig);
    push_onto(context(0), new DefineSignal(clone(sig)));
    clone.clearSig(sig);
  }
  return Status();
}
#line 1821 "ASTGRC.nw"
Status Surface::visit(Var &s) {
  context(0) = synthesize(s.body);
  return Status();
}
#line 2118 "ASTGRC.nw"
  
#line 736 "ASTGRC.nw"
Status Depth::visit(Pause &s) {
  return Status();
}
#line 865 "ASTGRC.nw"
Status Depth::visit(IfThenElse &s) {
  Switch *sw = new Switch(stnode(s));
  *sw >> ( (s.else_part != 0) ? recurse(s.else_part) : context(0))
      >> ( (s.then_part != 0) ? recurse(s.then_part) : context(0));
  context(0) = sw;
  return Status();
}
#line 925 "ASTGRC.nw"
Status Depth::visit(StatementList &s) {
  Switch *sw;
  if (!s.statements.empty()) {
    sw = new Switch(stnode(s));
    environment.surface_context.push(context);
    vector<Statement*>::reverse_iterator final = s.statements.rend();
    final--;
    for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	  i != s.statements.rend() ; i++ ) {
      assert(*i);
      *sw >> synthesize(*i); // Build the depth
      // Build the surface
      if (i != final ) context(0) = environment.surface.synthesize(*i);
    }
    environment.surface_context.pop();
    context(0) = sw;
  }
  return Status();
}
#line 983 "ASTGRC.nw"
Status Depth::visit(Loop &s) {
  environment.surface_context.push(context);
  // Synthesize the surface
  context(0) = environment.surface.synthesize(s.body);
  // Synthesize the depth
  context(0) = synthesize(s.body);
  environment.surface_context.pop();
  return Status();
}
#line 1032 "ASTGRC.nw"
Status Depth::visit(Repeat &s) {

  // std::cerr<<"depth visit\n";

  // Synthesize the surface
  environment.surface_context.push(context);
  GRCNode *restart = environment.surface.synthesize(s.body);
  environment.surface_context.pop();

  Test *tst = new Test(stnode(s), new CheckCounter(environment.boolean_type, s.counter));
  *tst >> restart >> context(0);
  //Synthesize the depth
  context(0) = tst;
  context(0) = synthesize(s.body);
 
  // std::cerr<<"visit ok\n";

  return Status();
}
#line 1126 "ASTGRC.nw"
Status Depth::visit(Suspend &s) {

  assert(s.predicate);
  assert(s.body);
  assert(stnode(s));

  assert(stnode(s)->children.size() >= 1); // Should have at least one child
  STNode *bodytree = stnode(s)->children.front();
  assert(bodytree); // First child is STref for the body

  Switch *swimm = new Switch(stnode(s));

  GRCNode *start = synthesize(s.body);

  Delay *d = dynamic_cast<Delay*>(s.predicate);
  Expression *pred =
    (d != NULL) ?
      ( d->is_immediate ?
	clone(d->predicate) :
	new CheckCounter(environment.boolean_type, d->counter)
      ) :
    clone(s.predicate);

  // the depth test: body is already started
  
  Test *t = new Test(bodytree, pred);
  STSuspend *sts = new STSuspend(bodytree);
  *sts >> context(1);
  *t >> start >> sts;
  Enter *hold = new Enter(bodytree);
  *hold >> t;
  *swimm >> hold;

  // If the predicate is immediate then it's possible that the surface
  // still needs to start in the depth of the suspend (i.e., when
  // it was suspended in the first cycle)
  //
  // This section builds that portion of the code

  if (d && d->is_immediate) {
    assert(stnode(s)->children.size() == 2); // build in selection tree
    Enter *enimleaf = new Enter( stnode(s)->children.back() );
    Enter *en = new Enter( bodytree );
    *enimleaf >> context(1);
    t = new Test(NULL, clone(pred));
    start = environment.surface.recurse(s.body);
    push_onto(start, en);
    *t >> start >>enimleaf;
    *swimm >> t;
  }
 
  context(0) = swimm;
  return Status();
}
#line 1311 "ASTGRC.nw"
Status Depth::visit(Abort &s) {
  if (s.is_weak) throw IR::Error("weak abort.  Did the dismantler run?");

  // Synthesize the depth of the body

  context.push();
  assert(s.body);
  GRCNode *resume = recurse(s.body);//
  context.pop();

  // The selection tree node for the body of the abort

  STNode *bodytree = stnode(s)->children.front();
  push_onto(resume, new Enter(bodytree));

  // Add a check for each predicate that branches to the surface of
  // the handler.  Also add the depth of each handler

  for ( vector<PredicatedStatement*>::reverse_iterator i = s.cases.rbegin() ;
	i != s.cases.rend() ; i++ ) {
    assert(*i);
    assert((*i)->predicate);

    // Get the predicate expression: either the simple predicate,
    // the predicate of an immediate, or a counter check for counted
    // predicates

    Delay *d = dynamic_cast<Delay*>((*i)->predicate);
    Expression *pred =
      d ?
      (d->is_immediate ?
       clone(d->predicate) : new CheckCounter(environment.boolean_type, d->counter))
      :  clone((*i)->predicate);

    // Add a test for the predicate

    Test *tst = new Test( bodytree, pred );
    GRCNode *handler;
    if ((*i)->body) {
      environment.surface_context.push(context);
      handler = environment.surface.synthesize((*i)->body);
      environment.surface_context.pop();
    } else handler = context(0);

    *tst >> resume >> handler;
    resume = tst;
  }
 
  // The switch at the top of the depth for the abort

  Switch *topswitch = new Switch( stnode(s) );

  // Its first child is the code for the body of the abort

  *topswitch >> resume;

  // Its remaining children are the bodies of the handlers

  for ( vector<PredicatedStatement*>::const_iterator i = s.cases.begin() ;
	i != s.cases.end() ; i++ ) {
    assert(*i);
    assert((*i)->predicate);
    if ((*i)->body) *topswitch >> recurse((*i)->body);
  }

  context(0) = topswitch;

  return Status();
}
#line 1480 "ASTGRC.nw"
Status Depth::visit(ParallelStatementList &s) {
  Fork *fork = new Fork();
  Sync *sync = new Sync(stnode(s));
  Enter *en;
  int nthr;
  Terminate *t;

  GRCNode **outer = context.continuations.top();
  assert(outer);
  context.push();

  // Create a new terminate for every possible exit level
  // and link each from the sync node


  // Synthesize each thread's surface
  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++) {
    assert(*i);
    nthr=i-s.threads.begin();
    Switch *sw = new Switch( stnode(s)->children[nthr] );
    assert(sw->st);
    *fork >> sw;
 
    for(int tl = 0; tl < context.size; tl++){
	t = new Terminate(tl, nthr);
	context(tl)=t;
	*t >> sync;
	if(tl == 0){
	 // this is the self looping enter
          en=new Enter( stnode(s)->children[nthr]->children[0] );
          assert(en->st);
          *sw >> en;
          push_onto(context(tl), en);	
	}
    }

   *sw >> recurse(*i);
  } 

  // Connect each Terminate node with predecessors (i.e., that was
  // used by the threads) to the Sync and delete the rest.

  for ( int i = 0 ; i < context.size ; i++ )
	*sync >> outer[i];

  context.pop();
  context(0) = fork;
  push_onto(context(0), new Enter(stnode(s))); // hold
  return Status();
}
#line 1659 "ASTGRC.nw"
Status Depth::visit(Trap &s) {

  assert(s.symbols);
  assert(s.symbols->begin() != s.symbols->end());
  SignalSymbol *ts = dynamic_cast<SignalSymbol*>((*(s.symbols->begin())));
  assert(ts);
  int level = environment.code[ts];
  assert(level > 1); // Should have been assigned by the dismantler

  // Clone each of the trap signals

  for (SymbolTable::const_iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++ ) {
    SignalSymbol *ss = dynamic_cast<SignalSymbol*>(*i);
    assert(ss);
    assert(ss->kind == SignalSymbol::Trap);
    clone.cloneLocalSignal(ss, environment.module->signals);
  }

  // Esterel permits multiple handlers, but the dismantler should
  // have removed them
  if (s.handlers.size() > 1)
    throw IR::Error("Multiple trap handler.  Did the dismantler run?");

  // Surface for the handler is either a normal termination or
  // the handler for the body
  assert( s.handlers.empty() || s.handlers.front() );
  GRCNode *handlerSurface = context(0);
  if (!s.handlers.empty()) {
    environment.surface_context.push(context);   
    handlerSurface = environment.surface.synthesize(s.handlers.front()->body);
    environment.surface_context.pop();
  }

  GRCNode *handlerDepth =
     s.handlers.empty() ? 0 : recurse(s.handlers.front()->body);

  // Synthesize the body
  context.push();

  assert(handlerSurface);
  context(level) = handlerSurface;

  assert(s.body);
  GRCNode *depth = synthesize(s.body);

  context.pop();

  Switch *topswitch = new Switch( stnode(s) );
  *topswitch >> depth;

  if (handlerDepth) *topswitch >> handlerDepth;

  // Delete the mapping for each of the traps
  for (SymbolTable::const_iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++ ) {
    SignalSymbol *ss = dynamic_cast<SignalSymbol*>(*i);
    assert(ss);
    assert(ss->kind == SignalSymbol::Trap);
    clone.clearSig(ss);
  } 

  context(0) = topswitch;
  return Status();
}
#line 1775 "ASTGRC.nw"
Status Depth::visit(Signal &s) {
  for ( SymbolTable::const_iterator i = s.symbols->begin() ;
        i != s.symbols->end() ; i++ ) {
    SignalSymbol *sig = dynamic_cast<SignalSymbol*>(*i);
    assert(sig);
    clone.cloneLocalSignal(sig, environment.module->signals);
  }

  context(0) = synthesize(s.body);
   for ( SymbolTable::const_iterator i = s.symbols->begin() ;
	 i!= s.symbols->end() ; i++ ) {
     SignalSymbol *sig = dynamic_cast<SignalSymbol*>(*i);
     assert(sig);
     push_onto(context(0), new DefineSignal(clone(sig)));
     clone.clearSig(sig);
  }
  return Status();
}
#line 1828 "ASTGRC.nw"
Status Depth::visit(Var &s) {
  context(0) = synthesize(s.body);
  return Status();
}
#line 2119 "ASTGRC.nw"
  
#line 667 "ASTGRC.nw"
void SelTree::setNode(const ASTNode &n, STNode *sn) {
  assert(sn);
  environment.ast2st[&n] = sn;
}
#line 706 "ASTGRC.nw"
Status SelTree::visit(Pause &s){
  STleaf *leaf = new STleaf();
  setNode(s, leaf);
  return Status(leaf);
}
#line 826 "ASTGRC.nw"
Status SelTree::visit(IfThenElse &s) {  
  STexcl *ite = new STexcl();
  setNode(s, ite);

  *ite >> (s.else_part ? synthesize(s.else_part) : new STref())
       >> (s.then_part ? synthesize(s.then_part) : new STref());

  return Status(ite);
}
#line 885 "ASTGRC.nw"
Status  SelTree::visit(StatementList &s)
{
  STexcl *excl = new STexcl();
  setNode(s, excl);

  for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	i != s.statements.rend() ; i++ ){
    assert(*i);
    *excl >> synthesize(*i);
  }

  return Status(excl);
}


#line 957 "ASTGRC.nw"
Status SelTree::visit(Loop &s) {
  STref *lp = new STref();
  setNode(s, lp);
  *lp >> synthesize(s.body);
  return Status(lp);
}
#line 1004 "ASTGRC.nw"
Status SelTree::visit(Repeat &s) {
  STref *lp = new STref();
  setNode(s, lp);
  *lp >> synthesize(s.body);
  return Status(lp);
}
#line 1065 "ASTGRC.nw"
Status SelTree::visit(Suspend &s)
{
  STexcl *ex = new STexcl();
  STref *sp = new STref();
  sp->setsuspend();
 
  *ex >> sp;
  Delay *d = dynamic_cast<Delay*>(s.predicate);
  if (d && d->is_immediate) *ex >> new STleaf();

  assert(s.body);
  *sp >> synthesize(s.body);

  setNode(s, ex);
  return Status(ex);
}
#line 1203 "ASTGRC.nw"
Status SelTree::visit(Abort &s) {

  // The selection tree for the body of the abort:
  //  An STref node whose only child is the tree
  //  for the body of the abort

  assert(s.body); // Any abort should have a body
  STref *bodytree = new STref();
  bodytree->setabort();
  *bodytree >> synthesize(s.body);

  // The root of the tree for the abort: an exclusive
  // whose first child is the tree for the body of the abort

  STexcl *exclusive = new STexcl();
  *exclusive >> bodytree;

  // Attach the selection tree for each non-vacuous handler
  // under the exclusive node at the top of the abort

  for ( vector<PredicatedStatement*>::const_iterator i = s.cases.begin() ;
	  i != s.cases.end() ; i++ ) {
      assert(*i);
      assert((*i)->predicate);
      if ((*i)->body) *exclusive >> synthesize((*i)->body);
  }

  setNode(s, exclusive);
  return Status(exclusive);
}
#line 1403 "ASTGRC.nw"
Status SelTree::visit(ParallelStatementList &s)
{
  STpar *par = new STpar();
  setNode(s, par);

  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++ ) {
    assert(*i);
    STexcl *ex = new STexcl();
    *par >> ex; 
    STleaf *term = new STleaf();
    term->setfinal();
    *ex >> term;
    *ex >> synthesize(*i);
  }

  return Status(par);
}
#line 1546 "ASTGRC.nw"
Status SelTree::visit(Trap &s) {

  // Create the topmost exclusive node

  STexcl *exclusive = new STexcl();

  // Create the subtree for the body of the Trap; attach it to the top

  assert(s.body);
  STref *bodytree = new STref();
  *bodytree >> synthesize(s.body);
  *exclusive >> bodytree;

  // Create the subtree for the handler, if any

  switch (s.handlers.size()) {
  case 0:
    // No handler; nothing to do
    break;
  case 1:
    // Single handler: add the selection tree for it to the exclusive node
    assert(s.handlers.front());
    assert(s.handlers.front()->body);
    *exclusive >> synthesize(s.handlers.front()->body);
    break;
  default:
    // Esterel permits multiple handlers, but the dismantler should
    // have removed them
    throw IR::Error("Multiple trap handler.  Did the dismantler run?");
    break;
  }

  setNode(s, exclusive);
  return Status(exclusive);  
}
#line 1745 "ASTGRC.nw"
Status SelTree::visit(Signal &s) {
  STNode *st = new STref();
  *st >> synthesize(s.body);
  return Status(st);
}
#line 1813 "ASTGRC.nw"
Status SelTree::visit(Var &s) {
  STNode *st = new STref();
  *st >> synthesize(s.body);
  return Status(st);
}
#line 2120 "ASTGRC.nw"
  
#line 1904 "ASTGRC.nw"
void Dependencies::compute(GRCNode *root)
{
  assert(root);

  Dependencies depper;

  depper.dfs(root);

  for ( map<SignalSymbol *, SignalNodes>::const_iterator i =
	  depper.dependencies.begin() ; i != depper.dependencies.end() ;
	i++ ) {
    const SignalNodes &sn = (*i).second;
    if (!sn.writers.empty() && !sn.readers.empty()) {
      for ( set<GRCNode*>::const_iterator j = sn.writers.begin() ;
	    j != sn.writers.end() ; j++ )
	for ( set<GRCNode*>::const_iterator k = sn.readers.begin() ;
	      k != sn.readers.end() ; k++ )
	  **k << *j;
    }
  }
}
#line 1934 "ASTGRC.nw"
void Dependencies::dfs(GRCNode *n)
{
  if (!n || visited.find(n) != visited.end() ) return;

  visited.insert(n);

  current = n;
  n->welcome(*this);

  for (vector<GRCNode*>::const_iterator i = n->successors.begin() ;
       i < n->successors.end() ; i++ ) dfs(*i);  
}
#line 1957 "ASTGRC.nw"
Status Dependencies::visit(Action &act)
{
    Emit *emt = dynamic_cast<Emit *>(act.body);
    if (emt) {
      dependencies[emt->signal].writers.insert(current);
    } else {
      // Traps are treated the same as signals
      Exit *ex = dynamic_cast<Exit *>(act.body);
      if (ex) dependencies[ex->trap].writers.insert(current);
    }
    return Status();
}
#line 2043 "ASTGRC.nw"
Status Dependencies::visit(Sync &s)
{
  for ( vector<GRCNode*>::const_iterator i = s.predecessors.begin() ;
        i != s.predecessors.end() ; i++ ) {
    // Every predecessor should be a terminate node
    assert( dynamic_cast<Terminate*>(*i) );
    s << *i;
  }
  return Status();
}
#line 2121 "ASTGRC.nw"
}
